import ctypes
import logging
import unittest

from spectator import Config, MemoryWriter, Registry


class RegistryTest(unittest.TestCase):

    def test_close(self):
        r = Registry(Config("memory"))

        c = r.counter("counter")
        c.increment()
        self.assertEqual("c:counter:1", r.writer().last_line())

        r.close()
        self.assertTrue(r.writer().is_empty())

    def test_age_gauge(self):
        r = Registry(Config("memory"))

        g1 = r.age_gauge("age_gauge")
        g2 = r.age_gauge("age_gauge", {"my-tags": "bar"})
        self.assertTrue(r.writer().is_empty())

        g1.set(1)
        self.assertEqual("A:age_gauge:1", r.writer().last_line())

        g2.set(2)
        self.assertEqual("A:age_gauge,my-tags=bar:2", r.writer().last_line())

    def test_age_gauge_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        g = r.age_gauge_with_id(r.new_id("age_gauge", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        g.set(0)
        self.assertEqual("A:age_gauge,extra-tags=foo,my-tags=bar:0", r.writer().last_line())

    def test_counter(self):
        r = Registry(Config("memory"))

        c1 = r.counter("counter")
        c2 = r.counter("counter", {"my-tags": "bar"})
        self.assertTrue(r.writer().is_empty())

        c1.increment()
        self.assertEqual("c:counter:1", r.writer().last_line())

        c2.increment()
        self.assertEqual("c:counter,my-tags=bar:1", r.writer().last_line())

        c1.increment(2)
        self.assertEqual("c:counter:2", r.writer().last_line())

        c2.increment(2)
        self.assertEqual("c:counter,my-tags=bar:2", r.writer().last_line())

        r.counter("counter").increment(3)
        self.assertEqual("c:counter:3", r.writer().last_line())

    def test_counter_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        c = r.counter_with_id(r.new_id("counter", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        c.increment()
        self.assertEqual("c:counter,extra-tags=foo,my-tags=bar:1", r.writer().last_line())

        c.increment(2)
        self.assertEqual("c:counter,extra-tags=foo,my-tags=bar:2", r.writer().last_line())

        r.counter("counter", {"my-tags": "bar"}).increment(3)
        self.assertEqual("c:counter,extra-tags=foo,my-tags=bar:3", r.writer().last_line())

    def test_distribution_summary(self):
        r = Registry(Config("memory"))

        d = r.distribution_summary("distribution_summary")
        self.assertTrue(r.writer().is_empty())

        d.record(42)
        self.assertEqual("d:distribution_summary:42", r.writer().last_line())

    def test_distribution_summary_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        d = r.distribution_summary_with_id(r.new_id("distribution_summary", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        d.record(42)
        self.assertEqual("d:distribution_summary,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_gauge(self):
        r = Registry(Config("memory"))

        g = r.gauge("gauge")
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("g:gauge:42", r.writer().last_line())

    def test_gauge_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        g = r.gauge_with_id(r.new_id("gauge", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("g:gauge,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_gauge_with_id_with_ttl_seconds(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        g = r.gauge_with_id(r.new_id("gauge", {"my-tags": "bar"}), ttl_seconds=120)
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("g,120:gauge,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_gauge_with_ttl_seconds(self):
        r = Registry(Config("memory"))

        g = r.gauge("gauge", ttl_seconds=120)
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("g,120:gauge:42", r.writer().last_line())

    def test_max_gauge(self):
        r = Registry(Config("memory"))

        g = r.max_gauge("max_gauge")
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("m:max_gauge:42", r.writer().last_line())

    def test_max_gauge_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        g = r.max_gauge_with_id(r.new_id("max_gauge", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        g.set(42)
        self.assertEqual("m:max_gauge,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_monotonic_counter(self):
        r = Registry(Config("memory"))

        c = r.monotonic_counter("monotonic_counter")
        self.assertTrue(r.writer().is_empty())

        c.set(42)
        self.assertEqual("C:monotonic_counter:42", r.writer().last_line())

    def test_monotonic_counter_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        c = r.monotonic_counter_with_id(r.new_id("monotonic_counter", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        c.set(42)
        self.assertEqual("C:monotonic_counter,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_monotonic_counter_uint(self):
        r = Registry(Config("memory"))

        c = r.monotonic_counter_uint("monotonic_counter_uint")
        self.assertTrue(r.writer().is_empty())

        c.set(ctypes.c_uint64(42))
        self.assertEqual("U:monotonic_counter_uint:42", r.writer().last_line())

    def test_monotonic_counter_uint_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        c = r.monotonic_counter_uint_with_id(r.new_id("monotonic_counter_uint", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        c.set(ctypes.c_uint64(42))
        self.assertEqual("U:monotonic_counter_uint,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_new_id(self):
        r1 = Registry(Config("memory"))
        id1 = r1.new_id("id")
        self.assertEqual("MeterId(name=id, tags={})", str(id1))

        r2 = Registry(Config("memory", {"extra-tags": "foo"}))
        id2 = r2.new_id("id")
        self.assertEqual("MeterId(name=id, tags={'extra-tags': 'foo'})", str(id2))

    def test_pct_distribution_summary(self):
        r = Registry(Config("memory"))

        d = r.pct_distribution_summary("pct_distribution_summary")
        self.assertTrue(r.writer().is_empty())

        d.record(42)
        self.assertEqual("D:pct_distribution_summary:42", r.writer().last_line())

    def test_pct_distribution_summary_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        d = r.pct_distribution_summary_with_id(r.new_id("pct_distribution_summary", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        d.record(42)
        self.assertEqual("D:pct_distribution_summary,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_pct_timer(self):
        r = Registry(Config("memory"))

        t = r.pct_timer("pct_timer")
        self.assertTrue(r.writer().is_empty())

        t.record(42)
        self.assertEqual("T:pct_timer:42", r.writer().last_line())

    def test_pct_timer_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        t = r.pct_timer_with_id(r.new_id("pct_timer", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        t.record(42)
        self.assertEqual("T:pct_timer,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_timer(self):
        r = Registry(Config("memory"))

        t = r.timer("timer")
        self.assertTrue(r.writer().is_empty())

        t.record(42)
        self.assertEqual("t:timer:42", r.writer().last_line())

    def test_timer_with_id(self):
        r = Registry(Config("memory", {"extra-tags": "foo"}))

        t = r.timer_with_id(r.new_id("timer", {"my-tags": "bar"}))
        self.assertTrue(r.writer().is_empty())

        t.record(42)
        self.assertEqual("t:timer,extra-tags=foo,my-tags=bar:42", r.writer().last_line())

    def test_writer(self):
        r = Registry(Config("memory"))
        self.assertTrue(isinstance(r.writer(), MemoryWriter))

    def test_writer_debug_logging(self):
        logging.getLogger("spectator.writer").setLevel(logging.DEBUG)
        r = Registry(Config("memory"))
        self.assertTrue(r.writer().is_empty())

        expected_messages = [
            "DEBUG:spectator.writer:write line=c:counter:1",
        ]

        with self.assertLogs("spectator.writer", level='DEBUG') as logs:
            c = r.counter("counter")
            c.increment()

        self.assertEqual(expected_messages, logs.output)



import abc

from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion


class Meter(metaclass=abc.ABCMeta):
    def __init__(self, meter_id: MeterId, writer: WriterUnion, meter_type_symbol: str) -> None:
        self._id = meter_id
        self._meter_type_symbol = meter_type_symbol
        self._writer = writer

    def writer(self) -> WriterUnion:
        return self._writer



import logging
from typing import Optional

from spectator.config import Config
from spectator.meter.age_gauge import AgeGauge
from spectator.meter.counter import Counter
from spectator.meter.dist_summary import DistributionSummary
from spectator.meter.gauge import Gauge
from spectator.meter.max_gauge import MaxGauge
from spectator.meter.meter_id import MeterId
from spectator.meter.monotonic_counter import MonotonicCounter
from spectator.meter.monotonic_counter_uint import MonotonicCounterUint
from spectator.meter.percentile_dist_summary import PercentileDistributionSummary
from spectator.meter.percentile_timer import PercentileTimer
from spectator.meter.timer import Timer
from spectator.writer.new_writer import new_writer, WriterUnion


class Registry:
    """Registry is the main entry point for interacting with the Spectator library."""

    def __init__(self, config: Config = Config()) -> None:
        self._logger = logging.getLogger(__name__)
        self._config = config
        if config.is_global:
            self._logger.debug("Create GlobalRegistry with extra_common_tags=%s",config.extra_common_tags)
        else:
            self._logger.info("Create Registry with extra_common_tags=%s",config.extra_common_tags)

        self._writer = new_writer(config)

    def writer(self) -> WriterUnion:
        return self._writer

    def close(self) -> None:
        self._logger.debug("Close Registry Writer")
        try:
            self._writer.close()
        except IOError:
            self._logger.error("Error closing Registry Writer")

    def new_id(self, name: str, tags: Optional[dict] = None) -> MeterId:
        """Create a new MeterId, which applies any configured extra common tags,
        and can be used as an input to the *_with_id Registry methods."""
        if tags is None:
            tags = {}

        new_meter_id = MeterId(name, tags)

        if len(self._config.extra_common_tags) > 0:
            new_meter_id = new_meter_id.with_tags(self._config.extra_common_tags)

        return new_meter_id

    def age_gauge(self, name: str, tags: Optional[dict] = None) -> AgeGauge:
        return AgeGauge(self.new_id(name, tags), self._writer)

    def age_gauge_with_id(self, meter_id: MeterId) -> AgeGauge:
        return AgeGauge(meter_id, self._writer)

    def counter(self, name: str, tags: Optional[dict] = None) -> Counter:
        return Counter(self.new_id(name, tags), self._writer)

    def counter_with_id(self, meter_id: MeterId) -> Counter:
        return Counter(meter_id, self._writer)

    def distribution_summary(self, name: str, tags: Optional[dict] = None) -> DistributionSummary:
        return DistributionSummary(self.new_id(name, tags), self._writer)

    def distribution_summary_with_id(self, meter_id: MeterId) -> DistributionSummary:
        return DistributionSummary(meter_id, self._writer)

    def gauge(self, name: str, tags: Optional[dict] = None, ttl_seconds: Optional[int] = None) -> Gauge:
        return Gauge(self.new_id(name, tags), self._writer, ttl_seconds)

    def gauge_with_id(self, meter_id: MeterId, ttl_seconds: Optional[int] = None) -> Gauge:
        return Gauge(meter_id, self._writer, ttl_seconds)

    def max_gauge(self, name: str, tags: Optional[dict] = None) -> MaxGauge:
        return MaxGauge(self.new_id(name, tags), self._writer)

    def max_gauge_with_id(self, meter_id: MeterId) -> MaxGauge:
        return MaxGauge(meter_id, self._writer)

    def monotonic_counter(self, name: str, tags: Optional[dict] = None) -> MonotonicCounter:
        return MonotonicCounter(self.new_id(name, tags), self._writer)

    def monotonic_counter_with_id(self, meter_id: MeterId) -> MonotonicCounter:
        return MonotonicCounter(meter_id, self._writer)

    def monotonic_counter_uint(self, name: str, tags: Optional[dict] = None) -> MonotonicCounterUint:
        return MonotonicCounterUint(self.new_id(name, tags), self._writer)

    def monotonic_counter_uint_with_id(self, meter_id: MeterId) -> MonotonicCounterUint:
        return MonotonicCounterUint(meter_id, self._writer)

    def pct_distribution_summary(self, name: str, tags: Optional[dict] = None) -> PercentileDistributionSummary:
        return PercentileDistributionSummary(self.new_id(name, tags), self._writer)

    def pct_distribution_summary_with_id(self, meter_id: MeterId) -> PercentileDistributionSummary:
        return PercentileDistributionSummary(meter_id, self._writer)

    def pct_timer(self, name: str, tags: Optional[dict] = None) -> PercentileTimer:
        return PercentileTimer(self.new_id(name, tags), self._writer)

    def pct_timer_with_id(self, meter_id: MeterId) -> PercentileTimer:
        return PercentileTimer(meter_id, self._writer)

    def timer(self, name: str, tags: Optional[dict] = None) -> Timer:
        return Timer(self.new_id(name, tags), self._writer)

    def timer_with_id(self, meter_id: MeterId) -> Timer:
        return Timer(meter_id, self._writer)



import os
import unittest
from typing import Dict

from spectator.common_tags import tags_from_env_vars, validate_tags


class CommonTagsTest(unittest.TestCase):

    @staticmethod
    def all_expected_tags() -> Dict[str, str]:
        return {
            "nf.container": "main",
            "nf.process": "python",
        }

    @staticmethod
    def setup_environment() -> None:
        os.environ["NETFLIX_PROCESS_NAME"] = "python"
        os.environ["TITUS_CONTAINER_NAME"] = "main"

    @staticmethod
    def clear_environment() -> None:
        keys = ["NETFLIX_PROCESS_NAME", "TITUS_CONTAINER_NAME"]

        for key in keys:
            try:
                del os.environ[key]
            except KeyError:
                pass

    def test_tags_from_env_vars(self):
        self.setup_environment()
        self.assertEqual(self.all_expected_tags(), tags_from_env_vars())
        self.clear_environment()

    def test_tags_from_env_vars_empty_ignored(self):
        self.setup_environment()
        os.environ["TITUS_CONTAINER_NAME"] = ""

        expected_tags = self.all_expected_tags()
        del expected_tags["nf.container"]
        self.assertEqual(expected_tags, tags_from_env_vars())

        self.clear_environment()

    def test_tags_from_env_vars_none_ignored(self):
        self.setup_environment()
        del os.environ["TITUS_CONTAINER_NAME"]

        expected_tags = self.all_expected_tags()
        del expected_tags["nf.container"]
        self.assertEqual(expected_tags, tags_from_env_vars())

        self.clear_environment()

    def test_tags_from_env_vars_whitespace_ignored(self):
        self.setup_environment()
        os.environ["TITUS_CONTAINER_NAME"] = "    main \t\t"
        self.assertEqual(self.all_expected_tags(), tags_from_env_vars())
        self.clear_environment()

    def test_validate_tags(self):
        self.assertEqual({}, validate_tags({"": "empty key", 1: "non-string key"}))
        self.assertEqual({}, validate_tags({"empty val": "", "non-string val": 1}))



import logging
import re
from typing import Dict, Optional

from spectator.common_tags import validate_tags


class MeterId:
    """The name and tags which uniquely identify a Meter instance. The tags are key-value pairs of
    strings. This class should NOT be used directly. Instead, use the Registry.new_id() method, to
    ensure that any extra common tags are properly applied to the Meter."""

    INVALID_CHARS = re.compile("[^-._A-Za-z0-9~^]")

    def __init__(self, name: str, tags: Optional[Dict[str, str]] = None) -> None:
        if tags is None:
            tags = {}
        self._logger = logging.getLogger(__name__)
        self._name = name
        self._tags = self._validate_tags(tags)
        self.spectatord_id = self._to_spectatord_id(self._name, self._tags)

    def _validate_tags(self, tags: Dict[str, str]) -> Dict[str, str]:
        valid_tags = validate_tags(tags)

        if valid_tags != tags:
            msg = "MeterId(name=%s, tags=%s) is invalid due to tag keys or values which are not strings " \
                  "or are zero-length strings; proceeding with truncated tags MeterId(name=%s, tags=%s)"
            self._logger.warning(msg, self._name, tags, self._name, valid_tags)

        return valid_tags

    def _replace_invalid_chars(self, s: str) -> str:
        return self.INVALID_CHARS.sub("_", s)

    def _to_spectatord_id(self, name: str, tags: Optional[Dict[str, str]]) -> str:
        if tags is None:
            tags = {}

        result = self._replace_invalid_chars(name)

        for k, v in sorted(tags.items()):
            k = self._replace_invalid_chars(k)
            v = self._replace_invalid_chars(v)
            result += f",{k}={v}"

        return result

    def name(self) -> str:
        return self._name

    def tags(self) -> Dict[str, str]:
        return self._tags.copy()

    def with_tag(self, k: str, v: str) -> "MeterId":
        new_tags = self._tags.copy()
        new_tags[k] = v
        return MeterId(self._name, new_tags)

    def with_tags(self, tags: Dict[str, str]) -> "MeterId":
        if len(tags) == 0:
            return self
        new_tags = self._tags.copy()
        new_tags.update(tags)
        return MeterId(self._name, new_tags)

    def __hash__(self):
        return hash((self._name, frozenset(self._tags.items())))

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __str__(self) -> str:
        return f"MeterId(name={self._name}, tags={self._tags})"



from spectator.writer import Writer
from typing import List


class MemoryWriter(Writer):
    """Writer that stores lines in a list, to support unit testing."""

    def __init__(self) -> None:
        super().__init__()
        self._logger.info("initialize MemoryWriter")
        self._messages: List[str] = []

    def write(self, line: str) -> None:
        self._logger.debug("write line=%s", line)
        self._messages.append(line)

    def close(self) -> None:
        self._messages.clear()

    def get(self) -> List[str]:
        return self._messages

    def clear(self) -> None:
        self._messages.clear()

    def is_empty(self) -> bool:
        return len(self._messages) == 0

    def last_line(self) -> str:
        return self._messages[-1]



from typing import List


class LineBuffer:
    """Line buffer for Writers."""

    def __init__(self, max_size: int) -> None:
        self._size: int = 0
        self._max_size: int = max_size
        self._lines: List[str] = []

    def __len__(self) -> int:
        return self._size

    def append(self, line: str) -> bool:
        """Return value indicates whether the buffer should be flushed."""
        self._lines.append(line)
        self._size += len(line)
        return self._size >= self._max_size

    def flush(self) -> str:
        lines = "\n".join(self._lines)
        self._lines = []
        self._size = 0
        return lines



import abc
import time


class Clock(metaclass=abc.ABCMeta):
    def wall_time(self) -> float:
        raise NotImplementedError

    def monotonic_time(self) -> float:
        raise NotImplementedError


class SystemClock(Clock):
    def wall_time(self) -> float:
        return time.time()

    def monotonic_time(self) -> float:
        """Use `time.perf_counter()`, because it is a clock with the highest available resolution
        to measure a short duration. It does include time elapsed during sleep and is system-wide.

        See https://docs.python.org/3/library/time.html#time.perf_counter for more information."""
        return time.perf_counter()


class ManualClock(Clock):
    """Used to enable deterministic unit testing for the StopWatch class."""
    def __init__(self, wall_init: float = 0, monotonic_init: float = 0) -> None:
        self._wall = wall_init
        self._monotonic = monotonic_init

    def wall_time(self) -> float:
        return self._wall

    def monotonic_time(self) -> float:
        return self._monotonic

    def set_wall_time(self, t: float) -> None:
        self._wall = t

    def set_monotonic_time(self, t: float) -> None:
        self._monotonic = t



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class AgeGauge(Meter):
    """The value is the time in seconds since the epoch at which an event has successfully
    occurred, or 0 to use the current time in epoch seconds. After an Age Gauge has been set,
    it will continue reporting the number of seconds since the last time recorded, for as long
    as the SpectatorD process runs. This meter type makes it easy to implement the Time Since
    Last Success alerting pattern."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "A")

    def now(self) -> None:
        line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:0"
        self._writer.write(line)

    def set(self, seconds: int) -> None:
        line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{seconds}"
        self._writer.write(line)



from typing import Optional, Union

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class Counter(Meter):
    """The value is the number of increments that have occurred since the last time it was
    recorded. The value will be reported to the Atlas backend as a rate-per-second."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "c")

    def increment(self, delta: Union[int, float] = 1) -> None:
        if delta > 0:
            line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{delta}"
            self._writer.write(line)



import os
import unittest
from typing import Dict

from spectator import Config


class ConfigTest(unittest.TestCase):

    @staticmethod
    def all_expected_tags() -> Dict[str, str]:
        return {
            "nf.container": "main",
            "nf.process": "python",
        }

    @staticmethod
    def setup_environment() -> None:
        os.environ["NETFLIX_PROCESS_NAME"] = "python"
        os.environ["TITUS_CONTAINER_NAME"] = "main"

    @staticmethod
    def clear_environment() -> None:
        keys = ["NETFLIX_PROCESS_NAME", "SPECTATOR_OUTPUT_LOCATION", "TITUS_CONTAINER_NAME"]

        for key in keys:
            try:
                del os.environ[key]
            except KeyError:
                pass

    @staticmethod
    def get_location(location: str) -> str:
        return Config(location).location

    def test_is_valid_output_location(self):
        config = Config()
        self.assertTrue(config.is_valid_output_location("none"))
        self.assertTrue(config.is_valid_output_location("memory"))
        self.assertTrue(config.is_valid_output_location("stdout"))
        self.assertTrue(config.is_valid_output_location("stderr"))
        self.assertTrue(config.is_valid_output_location("udp"))
        self.assertTrue(config.is_valid_output_location("unix"))
        self.assertTrue(config.is_valid_output_location("file://testfile.txt"))
        self.assertTrue(config.is_valid_output_location("udp://localhost:1234"))
        self.assertTrue(config.is_valid_output_location("unix:///tmp/socket.sock"))
        self.assertFalse(config.is_valid_output_location("invalid"))

    def test_default_config(self):
        self.setup_environment()
        config = Config()
        self.assertEqual(self.all_expected_tags(), config.extra_common_tags)
        self.assertEqual("udp", config.location)
        self.clear_environment()

    def test_env_location_override(self):
        os.environ["SPECTATOR_OUTPUT_LOCATION"] = "memory"
        config = Config()
        self.assertEqual("memory", config.location)
        self.clear_environment()

    def test_invalid_env_location_override(self):
        os.environ["SPECTATOR_OUTPUT_LOCATION"] = "invalid"
        self.assertRaises(ValueError, Config)
        self.clear_environment()

    def test_extra_common_tags(self):
        self.setup_environment()
        config = Config(extra_common_tags={"extra-tag": "foo"})
        self.assertEqual("udp", config.location)
        self.assertEqual({"extra-tag": "foo", "nf.container": "main", "nf.process": "python"}, config.extra_common_tags)
        self.clear_environment()

    def test_valid_output_locations(self):
        self.assertEqual("none", self.get_location("none"))
        self.assertEqual("memory", self.get_location("memory"))
        self.assertEqual("stderr", self.get_location("stderr"))
        self.assertEqual("stdout", self.get_location("stdout"))
        self.assertEqual("udp", self.get_location("udp"))
        self.assertEqual("unix", self.get_location("unix"))
        self.assertEqual("file://", self.get_location("file://"))
        self.assertEqual("udp://", self.get_location("udp://"))

    def test_invalid_output_locations(self):
        self.assertRaises(ValueError, self.get_location, None)
        self.assertRaises(ValueError, self.get_location, "foo")



from typing import Union

from spectator.config import Config
from spectator.writer.file_writer import FileWriter
from spectator.writer.memory_writer import MemoryWriter
from spectator.writer.noop_writer import NoopWriter
from spectator.writer.socket_writer import SocketWriter

WriterUnion = Union[FileWriter, MemoryWriter, NoopWriter, SocketWriter]


def new_writer(config: Config) -> WriterUnion:
    """Create a new Writer based on an output location."""

    if config.location == "none":
        writer: WriterUnion = NoopWriter()
    elif config.location == "memory":
        writer = MemoryWriter()
    elif config.location in ("stderr", "stdout"):
        writer = FileWriter(config)
    elif config.location == "udp":
        # default udp port for spectatord
        config.location = "udp://127.0.0.1:1234"
        writer = SocketWriter(config)
    elif config.location == "unix":
        # default unix domain socket for spectatord
        config.location = "unix:///run/spectatord/spectatord.unix"
        writer = SocketWriter(config)
    elif config.location.startswith("file://"):
        writer = FileWriter(config)
    elif config.location.startswith("udp://"):
        writer = SocketWriter(config)
    elif config.location.startswith("unix://"):
        writer = SocketWriter(config)
    else:
        raise ValueError(f"unsupported Writer location: {config.location}")

    return writer



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class DistributionSummary(Meter):
    """The value tracks the distribution of events. It is similar to a Timer, but more general,
    because the size does not have to be a period of time. For example, it can be used to
    measure the payload sizes of requests hitting a server or the number of records returned
    from a query."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "d")

    def record(self, amount: int) -> None:
        if amount >= 0:
            line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{amount}"
            self._writer.write(line)



import abc
import logging


class Writer(metaclass=abc.ABCMeta):
    def __init__(self):
        self._logger = logging.getLogger(__name__)

    @abc.abstractmethod
    def write(self, line: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError



import socket
import threading
import time
from ipaddress import ip_address, IPv6Address
from typing import Optional, Tuple, TypeAlias, Union
from urllib.parse import urlparse

from spectator.config import Config
from spectator.writer import Writer
from spectator.writer.line_buffer import LineBuffer

InetFamily: TypeAlias = Tuple[Union[str, None], Union[int, None]]
UnixFamily: TypeAlias = str
SocketAddress: TypeAlias = Union[InetFamily, UnixFamily]


class SocketWriter(Writer):
    """Writer that outputs data to either a UDP socket or a Unix Domain socket."""

    def __init__(self, config: Config) -> None:
        super().__init__()

        if config.is_global:
            self._logger.debug("initialize GlobalRegistry SocketWriter to %s (buffer_size=%s)",
                               config.location, config.buffer_size)
        else:
            self._logger.info("initialize SocketWriter to %s (buffer_size=%s)",
                              config.location, config.buffer_size)

        self._buffer: Optional[LineBuffer] = LineBuffer(config.buffer_size) if config.buffer_size > 0 else None
        self._lock = threading.Lock()
        self._sock: Optional[socket.socket] = None

        if "udp" in config.location:
            self._init_udp(config.location)
        else:
            self._init_unix(config.location)

        if self._buffer is not None:
            self._thread = threading.Thread(target=self._background_flush, daemon=True)
            self._logger.info("start SocketWriter background flush, every 5 seconds")
            self._thread.start()

    def _init_udp(self, location: str):
        parsed = urlparse(location)
        self._address: SocketAddress = (parsed.hostname, parsed.port)
        try:
            if self._address[0] is not None and type(ip_address(self._address[0])) is IPv6Address:
                self._family = socket.AF_INET6
            else:
                self._family = socket.AF_INET
        except ValueError:
            # anything that does not appear to be an IPv4 or IPv6 address (i.e. hostnames)
            self._family = socket.AF_INET

    def _init_unix(self, location: str):
        self._address = urlparse(location).path
        self._family = socket.AF_UNIX

    def _background_flush(self) -> None:
        while True:
            time.sleep(5)
            if self._buffer is None or self._sock is None:
                continue
            if len(self._buffer) > 0:
                with self._lock:
                    try:
                        self._sock.sendto(bytes(self._buffer.flush(), encoding="utf-8"), self._address)
                    except IOError:
                        self._logger.error("failed to write buffer from background flush")

    def _acquire_socket(self) -> None:
        # lazily instantiate the socket, in a thread-safe manner. this is necessary, because
        # the legacy GlobalRegistry will configure a SocketWriter (udp) upon `import spectator`.
        if self._sock is None:
            try:
                with self._lock:
                    if self._sock is None:
                        self._sock = socket.socket(family=self._family, type=socket.SOCK_DGRAM)
            except Exception as e:
                self._logger.error("exception during socket acquire: %s", e)

    def _write_buffer(self, line: str) -> None:
        if self._buffer is None or self._sock is None:
            return
        with self._lock:
            if self._buffer.append(line):
                try:
                    self._sock.sendto(bytes(self._buffer.flush(), encoding="utf-8"), self._address)
                except IOError:
                    self._logger.error("failed to write buffer, including line=%s", line)

    def _write_socket(self, line: str) -> None:
        if self._sock is None:
            return
        try:
            self._sock.sendto(bytes(line, encoding="utf-8"), self._address)
        except IOError:
            self._logger.error("failed to write line=%s", line)

    def write(self, line: str) -> None:
        self._logger.debug("write line=%s", line)
        self._acquire_socket()

        if self._buffer is not None:
            self._write_buffer(line)
        else:
            self._write_socket(line)

    def close(self) -> None:
        if self._sock is None:
            return
        self._sock.close()



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class MaxGauge(Meter):
    """The value is a number that was sampled at a point in time, but it is reported as a maximum
    gauge value to the backend."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "m")

    def set(self, value: float) -> None:
        line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{value}"
        self._writer.write(line)



from ctypes import c_uint64
from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class MonotonicCounterUint(Meter):
    """The value is a monotonically increasing number of an uint64 data type. These kinds of
    values are commonly seen in networking metrics, such as bytes-per-second. A minimum of two
    samples must be received in order for SpectatorD to calculate a delta value and report it to
    the backend."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "U")

    def set(self, amount: c_uint64) -> None:
        line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{amount.value}"
        self._writer.write(line)



from os import environ
from typing import Dict, Optional

from spectator.common_tags import tags_from_env_vars, validate_tags


class Config:
    """Create a new configuration with the provided location and extra common tags. All fields are
    optional. The extra common tags are added to every metric, on top of the common tags provided
    by `spectatord`.

    Possible values for `location` are:

      * `none`   - Configure a no-op writer that does nothing. Can be used to disable metrics collection.
      * `memory` - Write metrics to memory. Useful for testing.
      * `stderr` - Write metrics to standard error.
      * `stdout` - Write metrics to standard output.
      * `udp`    - Write metrics to the default spectatord UDP port. This is the default value.
      * `unix`   - Write metrics to the default spectatord Unix Domain Socket. Useful for high-volume scenarios.
      * `file:///path/to/file` - Write metrics to a file.
      * `udp://host:port`      - Write metrics to a UDP port.
      * `unix:///path/to/file` - Write metrics to a Unix Domain Socket.

    The output location can be overridden by configuring an environment variable `SPECTATOR_OUTPUT_LOCATION`
    with one of the values listed above. Overriding the output location may be useful for integration testing.

    The `buffer_size` is used to configure an optional `LineBuffer`, which caches protocol lines locally, before
    flushing them to `spectatord`. Flushes occur under two conditions: (1) the buffer size is exceeded, or (2)
    five seconds has elapsed. The buffer is available for the `SocketWriter` (udp and unix), where performance
    matters most, and it can increase the maximum RPS of communication to `spectatord` from ~80K to 150K (1.8X).
    The `LineBuffer` is disabled by default (with size zero), to ensure that the default operation of the library
    works under most circumstances. Under single-threaded performance testing, a 2KB buffer is a good configuration.

    When the `LineBuffer` is enabled, a background thread is started which will flush metrics every five seconds,
    which means that some care should be exercised when a custom Registry with buffering is instantiated in pre-fork
    environments.

    The `is_global` flag is used to help control logging for the GlobalRegistry.
    """

    def __init__(self, location: str = "udp", extra_common_tags: Optional[Dict[str, str]] = None,
                 buffer_size: int = 0, is_global: bool = False) -> None:
        self.location = self.calculate_location(location)
        if extra_common_tags is None:
            extra_common_tags = {}
        self.extra_common_tags = self.calculate_extra_common_tags(extra_common_tags)
        self.buffer_size = buffer_size
        self.is_global = is_global

    @staticmethod
    def is_valid_output_location(location: str) -> bool:
        if not isinstance(location, str):
            return False
        return location in ["none", "memory", "stdout", "stderr", "udp", "unix"] or \
            location.startswith("file://") or \
            location.startswith("udp://") or \
            location.startswith("unix://")

    def calculate_location(self, location: str) -> str:
        if not self.is_valid_output_location(location):
            raise ValueError(f"spectatord output location is invalid: {location}")

        override = environ.get("SPECTATOR_OUTPUT_LOCATION")
        if override is not None:
            if not self.is_valid_output_location(override):
                raise ValueError(f"SPECTATOR_OUTPUT_LOCATION is invalid: {override}")
            location = override

        return location

    @staticmethod
    def calculate_extra_common_tags(common_tags: Dict[str, str]) -> Dict[str, str]:
        merged_tags = validate_tags(common_tags)

        # merge common tags with env var tags; env vars take precedence
        for k, v in tags_from_env_vars().items():
            merged_tags[k] = v

        return merged_tags



from abc import ABCMeta
from typing import Optional, Tuple

from spectator.meter.age_gauge import AgeGauge
from spectator.meter.counter import Counter
from spectator.meter.dist_summary import DistributionSummary
from spectator.meter.gauge import Gauge
from spectator.meter.max_gauge import MaxGauge
from spectator.meter.meter_id import MeterId
from spectator.meter.monotonic_counter import MonotonicCounter
from spectator.meter.monotonic_counter_uint import MonotonicCounterUint
from spectator.meter.percentile_dist_summary import PercentileDistributionSummary
from spectator.meter.percentile_timer import PercentileTimer
from spectator.meter.timer import Timer

# https://netflix.github.io/atlas-docs/spectator/agent/usage/#metric-types
_METER_CLASSES = {
    'A': AgeGauge,
    'c': Counter,
    'd': DistributionSummary,
    'g': Gauge,
    'm': MaxGauge,
    'C': MonotonicCounter,
    'U': MonotonicCounterUint,
    'D': PercentileDistributionSummary,
    'T': PercentileTimer,
    't': Timer,
}


def get_meter_class(symbol: str) -> Optional[ABCMeta]:
    return _METER_CLASSES.get(symbol)


def parse_protocol_line(line: str) -> Tuple[str, MeterId, str]:
    """Parse a SpectatorD protocol line into component parts. Utility exposed for testing."""
    symbol, meter_id, value = line.split(":")
    # remove optional parts, such as gauge ttls
    symbol = symbol.split(",")[0]
    meter_id_parts = meter_id.split(",")
    name = meter_id_parts[0]

    tags = {}
    for tag in meter_id_parts[1:]:
        k, v = tag.split("=")
        tags[k] = v

    return symbol, MeterId(name, tags), value



import unittest

from spectator import Config, ManualClock, Registry, StopWatch


class StopwatchTest(unittest.TestCase):

    def test_stopwatch_with_pct_timer(self):
        r = Registry(Config("memory"))
        t = r.timer("pct_timer")
        self.assertTrue(r.writer().is_empty())

        clock = ManualClock()
        with StopWatch(t, clock):
            clock.set_monotonic_time(33)
        self.assertEqual("t:pct_timer:33", r.writer().last_line())

    def test_stopwatch_with_timer(self):
        r = Registry(Config("memory"))
        t = r.timer("timer")
        self.assertTrue(r.writer().is_empty())

        clock = ManualClock()
        with StopWatch(t, clock):
            clock.set_monotonic_time(42)
        self.assertEqual("t:timer:42", r.writer().last_line())



import unittest
from contextlib import closing

from spectator import GlobalRegistry
from .udp_server import UdpServer


class GlobalRegistryTest(unittest.TestCase):

    def test_counter(self):
        with closing(UdpServer(("127.0.0.1", 1234))) as server:
            c = GlobalRegistry.counter("counter")
            c.increment()
            self.assertEqual("c:counter:1", server.read())
            c.increment(2)
            self.assertEqual("c:counter:2", server.read())



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class Timer(Meter):
    """The value is the number of seconds that have elapsed for an event."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "t")

    def record(self, seconds: float) -> None:
        if seconds >= 0:
            line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{seconds}"
            self._writer.write(line)



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class Gauge(Meter):
    """The value is a number that was sampled at a point in time. The default time-to-live (TTL)
    for gauges is 900 seconds (15 minutes) - they will continue reporting the last value set for
    this duration of time. An optional ttl_seconds may be set to control the lifespan of these
    values. SpectatorD enforces a minimum TTL of 5 seconds."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None,
                 ttl_seconds: Optional[int] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        if ttl_seconds is None:
            meter_type_symbol = "g"
        else:
            meter_type_symbol = f"g,{ttl_seconds}"

        super().__init__(meter_id, writer, meter_type_symbol)

    def set(self, value: float) -> None:
        line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{value}"
        self._writer.write(line)



import sys
from urllib.parse import urlparse

from spectator.config import Config
from spectator.writer import Writer


class FileWriter(Writer):
    """Writer that outputs data to a TextIO instance, which can be stdout, stderr,
    or a regular file."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self._logger.info("initialize FileWriter to %s", config.location)

        if config.location == "stderr":
            self._file = sys.stderr
        elif config.location == "stdout":
            self._file = sys.stdout
        else:
            self._file = open(urlparse(config.location).path, "a", encoding="utf-8")

    def write(self, line: str) -> None:
        self._logger.debug("write line=%s", line)
        try:
            print(line, file=self._file)
        except IOError:
            self._logger.error("failed to write line=%s", line)

    def close(self) -> None:
        self._file.close()



from typing import Union

from spectator.clock import Clock, SystemClock
from spectator.meter.percentile_timer import PercentileTimer
from spectator.meter.timer import Timer


class StopWatch:
    """Context Manager that records the duration of a with-statement block in a Timer or
    PercentileTimer. It uses a custom Clock implementation to assist with unit testing."""

    def __init__(self, timer: Union[PercentileTimer, Timer], clock: Clock = SystemClock()) -> None:
        self._clock = clock
        self._timer = timer

    def __enter__(self) -> None:
        self._start = self._clock.monotonic_time()

    def __exit__(self, type_, value, traceback) -> None:
        self._timer.record(self.duration())

    def duration(self) -> float:
        return self._clock.monotonic_time() - self._start



from spectator.config import Config
from spectator.registry import Registry

from spectator.clock import Clock, SystemClock, ManualClock
from spectator.protocol_parser import get_meter_class, parse_protocol_line
from spectator.stopwatch import StopWatch

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId

from spectator.meter.age_gauge import AgeGauge
from spectator.meter.counter import Counter
from spectator.meter.dist_summary import DistributionSummary
from spectator.meter.gauge import Gauge
from spectator.meter.max_gauge import MaxGauge
from spectator.meter.monotonic_counter import MonotonicCounter
from spectator.meter.monotonic_counter_uint import MonotonicCounterUint
from spectator.meter.percentile_dist_summary import PercentileDistributionSummary
from spectator.meter.percentile_timer import PercentileTimer
from spectator.meter.timer import Timer

from spectator.writer import Writer
from spectator.writer.new_writer import new_writer, WriterUnion

from spectator.writer.file_writer import FileWriter
from spectator.writer.line_buffer import LineBuffer
from spectator.writer.memory_writer import MemoryWriter
from spectator.writer.noop_writer import NoopWriter
from spectator.writer.socket_writer import SocketWriter

# DEPRECATED: The GlobalRegistry construct is no longer necessary, since this library became a thin
# client implementation, but it was kept to help minimize the work associated with adopting the new
# version of this library. The previous advice for using this library offered many examples like the
# following:
#
#   from spectator import GlobalRegistry
#
#   GlobalRegistry.counter("server.numRequests").increment()
#
# Now, with the thin client version, this library is stateless. You can have one or more Registry
# objects in your code, and the preferred method of using it is as follows:
#
#   from spectator import Registry
#
#   r = Registry()
#   r.counter("server.numRequests").increment()
#
# Using this method of instantiating the Registry offers you the opportunity to provide an alternate
# configuration, to supply a different output location or a set of extra common tags. Or, keep the
# defaults and use the Registry as-is.

GlobalRegistry = Registry(Config(is_global=True))



import os

from typing import Dict


def _add_non_empty(tags: Dict[str, str], tag: str, *env_vars: str) -> None:
    for env_var in env_vars:
        value = os.environ.get(env_var)
        if value is None:
            continue
        value = value.strip()
        if len(value) != 0:
            tags[tag] = value
            break


def tags_from_env_vars() -> Dict[str, str]:
    """Extract common infrastructure tags from the Netflix environment variables, which are
    specific to a process and thus cannot be managed by a shared SpectatorD instance."""
    tags: Dict[str, str] = {}
    _add_non_empty(tags, "nf.container", "TITUS_CONTAINER_NAME")
    _add_non_empty(tags, "nf.process", "NETFLIX_PROCESS_NAME")
    return tags


def validate_tags(tags: Dict[str, str]) -> Dict[str, str]:
    valid_tags = {}

    for k, v in tags.items():
        if not isinstance(k, str) or len(k) == 0:
            continue
        if not isinstance(v, str) or len(v) == 0:
            continue
        valid_tags[k] = v

    return valid_tags



from spectator.writer import Writer


class NoopWriter(Writer):
    """Writer that does nothing. Used to disable output."""

    def __init__(self) -> None:
        super().__init__()
        self._logger.info("initialize NoopWriter")

    def write(self, line: str) -> None:
        self._logger.debug("write line=%s", line)

    def close(self) -> None:
        pass



import time
import unittest

from spectator import Clock, SystemClock, ManualClock


class ClockTest(unittest.TestCase):
    def test_wall_time_raises(self):
        self.assertRaises(NotImplementedError, Clock().wall_time)

    def test_monotonic_raises(self):
        self.assertRaises(NotImplementedError, Clock().monotonic_time)


class SystemClockTest(unittest.TestCase):
    def test_wall_time(self):
        c = SystemClock()
        start = c.wall_time()
        time.sleep(0.01)
        end = c.wall_time()
        self.assertTrue(end > start, "{} > {}".format(end, start))

    def test_monotonic_time(self):
        c = SystemClock()
        start = c.monotonic_time()
        time.sleep(0.01)
        end = c.monotonic_time()
        self.assertTrue(end > start, "{} > {}".format(end, start))


class ManualClockTest(unittest.TestCase):
    def test_init(self):
        c1 = ManualClock()
        self.assertEqual(0, c1.wall_time())
        self.assertEqual(0, c1.monotonic_time())

        c2 = ManualClock(1, 2)
        self.assertEqual(1, c2.wall_time())
        self.assertEqual(2, c2.monotonic_time())

    def test_set_wall_time(self):
        c = ManualClock()
        c.set_wall_time(1)
        self.assertEqual(1, c.wall_time())

    def test_set_monotonic_time(self):
        c = ManualClock()
        c.set_monotonic_time(1)
        self.assertEqual(1, c.monotonic_time())



from typing import Optional

from spectator.meter import Meter
from spectator.meter.meter_id import MeterId
from spectator.writer.new_writer import WriterUnion
from spectator.writer.noop_writer import NoopWriter


class PercentileDistributionSummary(Meter):
    """The value tracks the distribution of events. It is similar to a Timer, but more general,
    because the size does not have to be a period of time. For example, it can be used to
    measure the payload sizes of requests hitting a server or the number of records returned
    from a query.

    In order to maintain the data distribution, Percentile Distribution Summaries have a higher
    storage cost, with a worst-case of up to 300X that of a standard Distribution Summary. Be
    diligent about any additional dimensions added to Percentile Distribution Summaries and ensure
    that they have a small bounded cardinality."""

    def __init__(self, meter_id: MeterId, writer: Optional[WriterUnion] = None) -> None:
        if writer is None:
            writer = NoopWriter()

        super().__init__(meter_id, writer, "D")

    def record(self, amount: int) -> None:
        if amount >= 0:
            line = f"{self._meter_type_symbol}:{self._id.spectatord_id}:{amount}"
            self._writer.write(line)



import unittest

from spectator import Counter, Gauge, Timer, get_meter_class, parse_protocol_line


class ProtocolParserTest(unittest.TestCase):

    def test_parse_invalid_lines(self):
        self.assertRaises(ValueError, parse_protocol_line, "foo")
        self.assertRaises(ValueError, parse_protocol_line, "foo:bar")
        self.assertRaises(ValueError, parse_protocol_line, "foo:bar,baz-quux")

    def test_parse_counter(self):
        symbol, id, value = parse_protocol_line("c:counter:1")
        self.assertEqual("c", symbol)
        self.assertEqual(Counter, get_meter_class(symbol))
        self.assertEqual("counter", id.name())
        self.assertEqual({}, id.tags())
        self.assertEqual("1", value)

    def test_parse_counter_with_tag(self):
        symbol, id, value = parse_protocol_line("c:counter,foo=bar:1")
        self.assertEqual("c", symbol)
        self.assertEqual(Counter, get_meter_class(symbol))
        self.assertEqual("counter", id.name())
        self.assertEqual({"foo": "bar"}, id.tags())
        self.assertEqual("1", value)

    def test_parse_counter_with_multiple_tags(self):
        symbol, id, value = parse_protocol_line("c:counter,foo=bar,baz=quux:1")
        self.assertEqual("c", symbol)
        self.assertEqual(Counter, get_meter_class(symbol))
        self.assertEqual("counter", id.name())
        self.assertEqual({"foo": "bar", "baz": "quux"}, id.tags())
        self.assertEqual("1", value)

    def test_parse_gauge(self):
        symbol, id, value = parse_protocol_line("g:gauge:1")
        self.assertEqual("g", symbol)
        self.assertEqual(Gauge, get_meter_class(symbol))
        self.assertEqual("gauge", id.name())
        self.assertEqual({}, id.tags())
        self.assertEqual("1", value)

    def test_parse_gauge_with_ttl(self):
        symbol, id, value = parse_protocol_line("g,120:gauge:1")
        self.assertEqual("g", symbol)
        self.assertEqual(Gauge, get_meter_class(symbol))
        self.assertEqual("gauge", id.name())
        self.assertEqual({}, id.tags())
        self.assertEqual("1", value)

    def test_parse_timer(self):
        symbol, id, value = parse_protocol_line("t:timer:1")
        self.assertEqual("t", symbol)
        self.assertEqual(Timer, get_meter_class(symbol))
        self.assertEqual("timer", id.name())
        self.assertEqual({}, id.tags())
        self.assertEqual("1", value)
